# SPDX-FileCopyrightText: 2024 Brandenburgische Technische Universität Cottbus-Senftenberg
#
# SPDX-License-Identifier: EUPL-1.2

"""
PET plugin API wrapper.
"""

######################################################
# import commonly used libraries. You can add/remove
# libraries as needed
######################################################

import sys
import os
import json
import re
from pathlib import Path
from datetime import datetime
import subprocess

# from evaluation_system.api.exceptions import PluginError
from typing import List, Optional

from evaluation_system.api import plugin
from evaluation_system.api.parameters import (
    Directory,
    Integer,
    ParameterDictionary,
    File,
    SolrField,
    String,
    Bool,
    Date
    )
from evaluation_system.model.user import User
from evaluation_system.misc import logger

import freva

class PET(plugin.PluginAbstract):
    ######################################################
    # Define the details for the plugin
    ######################################################

    __version__ = (2025, 1, 28)
    __short_description__: str = (
        "Calculation of daily Potential Evatransporation"
    )
    __category__: Optional[str] = "postproc"
    __tags__: Optional[List[str]] = ["Potential Evatranspiration"]
    tool_developer = {
        "name": "Christian Beier",
        "email": "christian.beier@b-tu.de",
    }
    __long_description__: Optional[
        str
    ] = "This plugin calculates daily Potential Evapotranspiration (PET) based on the methodology outlined in FAO Irrigation and Drainage Paper No. 56 (Crop Evapotranspiration) by Allen, Pereira et al., 1998.\n\nFor more details, refer to the official publication: https://www.fao.org/4/x0490e/x0490e00.htm"

    ######################################################
    # Define the parameters for the plugin * Mandatory
    ######################################################

    __parameters__ = ParameterDictionary(
        SolrField(
            name="project",
            facet="project",
            mandatory=True,
            help="Project facet for input data search.",
            predefined_facets={
                "variable": ["hurs", "sfcWind"],
                "time_frequency": ["1day"],
                "project": ["cmip6", "cordex", "nukleus", "reanalysis"],
            }),
        SolrField(
            name="product",
            facet="product",
            mandatory=True,
            help="Product facet for input data search.",
            predefined_facets={
                "variable": ["hurs", "sfcWind"],
                "time_frequency": ["1day"],
            }),
        SolrField(
            name="institute",
            facet="institute",
            mandatory=True,
            help="Institute facet for input data search.",
            predefined_facets={
                "variable": ["hurs", "sfcWind"],
                "time_frequency": ["1day"],
            }),
        SolrField(
            name="model",
            facet="model",
            mandatory=True,
            help="Model facet for input data search.",
            predefined_facets={
                "variable": ["hurs", "sfcWind"],
                "time_frequency": ["1day"],
            }),
        SolrField(
            name="experiment",
            facet="experiment",
            mandatory=True,
            help="Experiment facet for input data search.",
            predefined_facets={
                "variable": ["hurs", "sfcWind"],
                "time_frequency": ["1day"],
            }),
        SolrField(
            name="ensemble",
            facet="ensemble",
            mandatory=True,
            help="ensemble facet for input data search.",
            predefined_facets={
                "variable": ["hurs", "sfcWind"],
                "time_frequency": ["1day"],
            }),
        Date(
            name="start_date",
            mandatory=False,
            default="1900-01-01",
            help="Insert start date you want to select (YYYY-MM-DD)"
        ),
        Date(
            name="end_date",
            mandatory=False,
            default="2999-01-01",
            help="Insert end date you want to select (YYYY-MM-DD)"
            ),
        String(
            name="region",
            default=None,
            help="Region box to be processed. The format have to be W,E,S,N (e.g. -180,180,0,90 for NH).\nLeave empty to select the region from input grid."
        ),
        Directory(
            name="output_dir",
            default="$USER_OUTPUT_DIR",
            help="Directory where output will be stored.",
        ),
        Bool(
            name="debug",
            default=False,
            help="Turn on to get better detailed debug information in case you face any issue."
        ),
    )

    # dictionary containing variable names for all projects (necessary because of different capitalization)
    project_vars = ["hurs", "tasmax", "tasmin", "sfcWind", "rsds"]

    def check_variables(self, config_dict):
        '''
        Checks if given set of facet values provides data for all required variables and suggestion more suitable options if this is not the case.

            Parameters:
                config_dict (dict): Dictionary containing facet values to search database
        '''

        facets = ["project", "product", "institute", "model", "experiment", "ensemble"]
        
        facet_suggestions = {}
        missing_vars = list(filter(lambda x: len(config_dict["files"][x]) == 0, config_dict["files"]))

        search_args = {key: config_dict[key] for key in facets}

        if (set(self.project_vars).issubset(missing_vars)):
            sys.exit(logger.error("There is no data for all required variables. Please check if the chosen data covers the defined time period."))
        elif (len(missing_vars) > 0):

            # remove facet stepwise and filter for possible alternative facet values
            removed_facets = []
            for facet in reversed(facets):
                del search_args[facet]
                removed_facets.append(facet)
                freva_results = dict(freva.facet_search(**search_args, facet="variable"))["variable"]

                # if alternative facets found covering all variables stop the loop and provide suggestion
                if (set(self.project_vars).issubset(freva_results)):

                    # check all possible entries of latest removed facet, all entries covering all variables will be saved as suggestions
                    facet_suggestions = []
                    if (facet == "project"):
                        facet_entries = ["cmip6", "cordex", "nukleus", "reanalysis"] # only search in relevant project datasets
                    else:
                        facet_entries = list(freva.facet_search(**search_args,facet=facet)[facet])

                    for sugg in facet_entries:
                        test_search_args = search_args.copy()
                        test_search_args[facet] = sugg
                        facet_variables = dict(freva.facet_search(**test_search_args, facet="variable"))["variable"]
                        
                        if (not set(self.project_vars).issubset(facet_variables)):
                            facet_suggestions.append(sugg)

                    # if at least one enrty covers all required variables, stop the search, otherwise continue
                    if (len(facet_suggestions) > 0):
                        error_msg = "There is data for some of the required variables missing (" + ", ".join(missing_vars) + ") ! Consider choose a different " + facet + ". \n Suggestions: " + ", ".join(facet_suggestions)
                        sys.exit(logger.error(error_msg))
                        break

    def find_orog(self, config_dict):
        '''
        Retrieves a list of orographic files based on given facets and returns first entry as suggestion

            Parameters:
                config_dict (dict): Dictionary containing facet values to search database
            Returns:
                orog_files[0] (string): First entry of list of found files
        '''
        facets = ["product", "institute", "model", "experiment", "ensemble"]
        search_args = {"project": config_dict["project"], "variable": "orog"}
        orog_files = list(freva.databrowser(**search_args))

        for facet in facets:
            search_args[facet] = config_dict[facet]
            file_number = freva.count_values(**search_args)
            
            # if there are results, keep them and the facet
            if (file_number > 0):
                orog_files = list(freva.databrowser(**search_args))
            else:
                del search_args[facet]

        return orog_files[0]

    def extract_timeperiod(self, config_dict):
        '''
        Retrieves start and end date based on given datasets and exchanges provided start and/or end date in configuration if data for users selected time period is not available

            Parameters:
                config_dict (dict): Dictionary containing facet values as well as user given time period
        '''

        # all variables should cover the same time period, we only need one to extract time period -> tasmax
        config_dict["files"]["tasmax"].sort(key=lambda f: int(''.join(filter(str.isdigit, f)))) #sorting list of files based on given start year in file name
       
        # cmorized daily data includes time period of data file as 'YYYYMMDD-YYYYMMDD' within the file name
        date1 = re.search(r'\d{4}\d{2}\d{2}-\d{4}\d{2}\d{2}', config_dict["files"]["tasmax"][0]).group(0).split("-",1)[0]
        date2 = re.search(r'\d{4}\d{2}\d{2}-\d{4}\d{2}\d{2}', config_dict["files"]["tasmax"][-1]).group(0).split("-",1)[1]

        start_date = datetime.strptime(date1, '%Y%m%d').strftime('%Y-%m-%d')
        end_date = datetime.strptime(date2, '%Y%m%d').strftime('%Y-%m-%d')

        time_adapted = False

        if (start_date > config_dict["start_date"]):
            config_dict["start_date"] = start_date
            time_adapted = True

        if (end_date < config_dict["end_date"]):
            config_dict["end_date"] = end_date
            time_adapted = True

        if (time_adapted):
            print('INFO Adapted time period based on data availability to ',config_dict["start_date"],'-', config_dict["end_date"])

    def find_files(self, config_dict, variables: List[str], time_frequency: str, time: Optional[str] = None):
        '''
        Selects data files based on the given parameters using the freva databrowser

            Parameters:
                config_dict (dict): Dictionary containing values used for searching the database
                variables (list): List of variables used to search data files within the database
                time_frequency (str): Abbreviation for time frequency the data should cover
                time (str): Time period the data should cover
            Returns:
                file_dict (dict): Dictionary containing list of files for each variable
        '''
        
        search_keys = (
            "project",
            "product",
            "institute",
            "model",
            "experiment",
            "ensemble"
        )

        search_args = {key: config_dict[key] for key in search_keys}
        search_args["time_frequency"] = time_frequency
        # TODO: instead of adding datetime manually as default, it would be ideal if the user had not added date, we calculate in start and end date based on the other facet selections and gathered files, OR always consider a big interval of datetime that all selected files can be included!
        search_args["time"] = time

        file_dict = {}
        for variable in variables:
            search_args["variable"] = variable
            files = list(freva.databrowser(**search_args))
            file_dict[variable] = files

        return file_dict

    def prepare_html_output(self, path_output):
        """
        Prepare the HTML output for the plugin in 
        the result section.
        """
        result = subprocess.run(
            ["ls", "-la", path_output], capture_output=True, text=True
        )
        ls_output = result.stdout
        current_user = User()

        html_content = f"""<!DOCTYPE html>
        <html>
        <head>
            <meta charset="UTF-8">
            <meta name="viewport" content="width=device-width, initial-scale=1.0, maximum-scale=1.0">
            <style>
                body {{
                    font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, sans-serif;
                    margin: 0;
                    padding: 20px;
                    line-height: 1.6;
                    background-color: white;
                    display: flex;
                    flex-direction: column;
                    align-items: center;
                    min-height: 100vh;
                }}
                .info-text {{
                    font-size: 16px;
                    max-width: 800px;
                    margin: 0 auto 20px auto;
                    color: #0c5460;
                    background-color: #d1ecf1;
                    border: 1px solid #bee5eb;
                    padding: 20px;
                    border-radius: 8px;
                    width: 100%;
                    box-sizing: border-box;
                }}
                .terminal {{
                    background-color: #2d2d2d;
                    border-radius: 8px;
                    box-shadow: 0 10px 20px rgba(0,0,0,0.3);
                    width: 100%;
                    max-width: 800px;
                    overflow: hidden;
                }}
                .terminal-header {{
                    background-color: #3d3d3d;
                    padding: 10px;
                    display: flex;
                    align-items: center;
                }}
                .terminal-buttons {{
                    display: flex;
                    gap: 8px;
                }}
                .terminal-button {{
                    width: 12px;
                    height: 12px;
                    border-radius: 50%;
                }}
                .terminal-button.close {{background-color: #ff5f56;}}
                .terminal-button.minimize {{background-color: #ffbd2e;}}
                .terminal-button.maximize {{background-color: #27c93f;}}
                .terminal-title {{
                    color: #ababab;
                    font-size: 14px;
                    margin-left: 20px;
                }}
                .terminal-body {{
                    padding: 20px;
                    font-family: "Menlo", "Monaco", "Consolas", monospace;
                    font-size: 14px;
                    line-height: 1.5;
                    color: #fff;
                    overflow-x: auto;
                    -webkit-overflow-scrolling: touch;
                }}
                .prompt {{
                    color: #98c379;
                    margin-bottom: 10px;
                    word-break: break-all;
                }}
                .command {{
                    color: #61afef;
                    margin-left: 8px;
                }}
                .output {{
                    color: #abb2bf;
                    white-space: pre-wrap;
                    margin: 10px 0 20px 0;
                }}
                @media (max-width: 480px) {{
                    body {{ padding: 10px; }}
                    .terminal-body {{ padding: 15px; }}
                }}                
            </style>
        </head>
        <body>
            <div class="info-text">
                You can directly download the produced data from the link above or login to levante and and proceed with the exploration of the data.
            </div>
            <div class="terminal">
                <div class="terminal-header">
                    <div class="terminal-buttons">
                        <div class="terminal-button close"></div>
                        <div class="terminal-button minimize"></div>
                        <div class="terminal-button maximize"></div>
                    </div>
                    <span class="terminal-title">Terminal</span>
                </div>
                <div class="terminal-body">
                    <div class="prompt">
                        local:~$ <span class="command">ssh {current_user.getName()}@levante.dkrz.de</span>
                    </div>
                    <div class="prompt">
                        {current_user.getName()}@levante.dkrz.de:~$ <span class="command">cd {path_output}</span>
                    </div>
                    <div class="prompt">
                        {current_user.getName()}@levante.dkrz.de:{path_output}$ <span class="command">ls -la {path_output}</span>
                    </div>
                    <div class="output">{ls_output}</div>
                </div>
            </div>
        </body>
        </html>"""

        html_output_path = os.path.join(path_output, "02_view.html")
        with open(html_output_path, "w") as f:
            f.write(html_content)

    def run_tool(self, config_dict=None):

        config_dict["files"] = self.find_files(config_dict, self.project_vars, "day", config_dict["start_date"] + " to " + config_dict["end_date"])
        config_dict["alti_file"] = self.find_orog(config_dict)
        if "alti_file" not in config_dict or not config_dict["alti_file"]:
            logger.error("Can't find altitude file")

        self.check_variables(config_dict)
        self.extract_timeperiod(config_dict)
        
        out_dir = self._special_variables.substitute({"d": "$USER_OUTPUT_DIR"})
        out_dir = Path(out_dir["d"]) / str(self.rowid)
        out_dir.mkdir(exist_ok=True, parents=True)
        config_dict["output_dir"] = str(out_dir)

        user = User()
        scratch = os.getenv("SCRATCH", f"/scratch/{user.getName()[0]}/{user.getName()}")
        cache_dir = Path(f"{scratch}/pet/{self.rowid}")
        cache_dir.mkdir(exist_ok=True, parents=True)
        config_dict["cache_dir"] = str(cache_dir)

        json_file = Path(config_dict["cache_dir"]) / "out.json"
        error_file = Path(config_dict["cache_dir"]) / "error.log"

        with json_file.open("w") as json_f:
            json.dump(config_dict, json_f, indent=4)

        tool_path = Path(__file__).parent / "src" / "pet" / "calculate_pet.sh"

        # TODO: adding a stage here to check if we have files in all keys in config_dict["files"] and config_dict["alti_file"] for better error handling as a predicted error:
        # logger.error("Your selected datetime doesn't included any data. Please change the datetime interval.")

        with error_file.open("w") as error_f:
            self.call(
                f"{tool_path} {json_file}",
                stderr=error_f,
            )
        self.prepare_html_output(config_dict["output_dir"])

        return self.prepare_output(config_dict["output_dir"])
