#!/bin/bash

# SPDX-FileCopyrightText: 2024 2024 Brandenburgische Technische Universität Cottbus-Senftenberg
#
# SPDX-License-Identifier: BSD-3-Clause

##################################################################################################
### LOAD MODULE + SYSTEM CONFIGURATION
##################################################################################################
# importing utilities
source $(dirname "$0")/utils.sh

if is_hpc_environment; then
    log_debug "HPC environment detected"
    load_hpc_modules
else
    log_debug "You are not on the HPC node. You have to install the required software packages (cdo, nco) manually."
fi

# importing functions
source $(dirname "$0")/pet_functions.sh

##################################################################################################
### PARAMETERS
##################################################################################################

### GENERAL VALUES
TMP_DIR=$(jq '.cache_dir' $1 | sed 's/\"//g')          # temporary directory
OUTPUT_DIR=$(jq '.output_dir' $1 | sed 's/\"//g')      # output directory
ALTI_FILE=$(jq '.alti_file' $1 | sed 's/\"//g')        # file containing altitude data

### INPUT DATA VALUES
PROJECT=$(jq '.project' $1 | sed 's/\"//g')
PRODUCT=$(jq '.product' $1 | sed 's/\"//g')
INSTITUTE=$(jq '.institute' $1 | sed 's/\"//g')
MODEL=$(jq '.model' $1 | sed 's/\"//g')
EXPERIMENT=$(jq '.experiment' $1 | sed 's/\"//g')
ENSEMBLE=$(jq '.ensemble' $1 | sed 's/\"//g')

### ANALYSIS VALUES
# todoo: test if time-period matches needed format
START_DATE=$(jq '.start_date' $1 | sed 's/\"//g')	# start date
END_DATE=$(jq '.end_date' $1 | sed 's/\"//g')           # end date
START_YEAR=$(cut -c 1-4 <<< ${START_DATE})		# start year
END_YEAR=$(cut -c 1-4 <<< ${END_DATE})			# end year
REGION=$(jq '.region' $1 | sed 's/\"//g')		# region (W,E,S,N)

### DEBUG MODE
DEBUG=$(jq '.debug' $1 | sed 's/\"//g')

### CONSTANTS ###
MAXPP=30

WIND_H_NAME=Wind_height  # variable name for height of wind speed (needed for log. wind profile)
PET_NAME=evspsblpot
PET_LNAME="potential daily evapotranspiration"
PET_SNAME="potential_daily_evapotranspiration"
PET_UNIT="kg m-2"

##################################################################################################
### OVERVIEW
##################################################################################################
echo ''
echo '### Chosen Input Parameters'
echo '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~'
echo ''
echo 'System Parameters:'
echo '------------------'
echo 'CACHE_DIR: '${TMP_DIR}
echo 'OUTPUT_DIR: '${OUTPUT_DIR}
echo 'ALTITUDE FILE: '${ALTI_FILE}
echo 'DEBUG: '${DEBUG}

echo ''
echo 'Input Parameters:'
echo '-----------------'
echo 'PROJECT: '${PROJECT}
echo 'PRODUCT: '${PRODUCT}
echo 'INSTITUTE: '${INSTITUTE}
echo 'MODEL: '${MODEL}
echo 'EXPERIMENT: '${EXPERIMENT}
echo 'ENSEMBLE: '${ENSEMBLE}
echo ''

echo 'Experiment Parameters'
echo '---------------------'
echo 'STAR DATE: '${START_DATE}
echo 'END DATE: '${END_DATE}
echo 'REGION: '${REGION}
echo '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~'
echo ''

##################################################################################################
### OUTPUT PREPARATION
##################################################################################################

echo ''
echo '### Preparing directories'
log_debug 'Creating output and cache directories...'
OUTPUT_DIR_VAR=${OUTPUT_DIR}'/'${PET_NAME}
[[ -d ${TMP_DIR} ]] || mkdir  ${TMP_DIR}   		# create temp. directory if not existing
[[ -d ${OUTPUT_DIR_VAR} ]] || mkdir  ${OUTPUT_DIR_VAR}  # create output directory if not existing
echo ''

##################################################################################################
### CHECKS
##################################################################################################

echo ''
echo '### Starting checks'
### CHECKING IF LIST OG DATA FILES CONTAINS ALL VARAIBLES NEEDED
log_debug 'Checking availability of all necessary variables...'

VARIABLES=$(jq '.files | keys | .[]' $1)
for VAR in ${VARIABLES[@]}; do
	if [ $(jq ".files.$VAR | length " $1) -gt 0 ]; then
		log_debug "${VAR//\"}...OK"
	else
		if [ ${VAR//\"} == "pvap" ]; then
			log_debug 'No data for '${VAR}' available. Data for '${VAR}' will be calculated.'
		else
			log_error 'ERROR: No data for '${VAR}' available!'
                	log_error 'Please select a dataset containg all necessary varaiables!'
                	exit 1
		fi
	fi
done

### PREPARING DATA (MERGING FILES; SELECTION TIME PERIOD) ###
echo ''
log_debug 'Preparing files for checks...'
for VAR in ${VARIABLES[@]}; do

  # skip preparation if no pvap data available
  if [ ${VAR//\"} == "pvap" ] && [ $(jq ".files.${VAR} | length" $1) == 0 ]; then
    log_debug "Skipping preparation of missing pvap data"
    continue
  fi

  # requirement: allfiles for a specific variable are located within one directory
  VAR_FILES=$(jq ".files.${VAR//\"}[]" $1 | sed 's/\"//g' | grep '.nc')
  TMP_FILE=${TMP_DIR}'/'${VAR//\"}'_'${START_YEAR}'-'${END_YEAR}'.nc'
  log_debug '---'${VAR//\"}'---'

  # merge files and select time period
  log_debug 'Selecting time period...'
  TIME_FILE=${TMP_DIR}'/'${VAR//\"}'_timeperiod.nc'
  if [ ! ${START_DATE} == null ] && [ ! ${END_DATE} == null ]; then
      cdo -s seldate,${START_DATE}'T00:00:00',${END_DATE}'T00:00:00' -mergetime ${VAR_FILES} ${TIME_FILE}
  fi

  if [ ! -f ${TIME_FILE} ]; then
      log_error 'Error selecting time and creating: '${TIME_FILE}
      exit 1
  else
      mv ${TIME_FILE} ${TMP_FILE}
  fi

  # selecting region
  if [ ! ${REGION} == null ]; then
      REGION_FILE=${TMP_DIR}'/'${VAR//\"}'_region.nc'

      log_debug 'Selecting region...'
      cdo -s -sellonlatbox,${REGION} ${TMP_FILE} ${REGION_FILE}

      if [ ! -f ${REGION_FILE} ]; then
          log_error "Error selecting region and creating: ${REGION_FILE}"
          exit 1
      else
          mv ${REGION_FILE} ${TMP_FILE}
      fi
  fi
done

### CHECKING VARIABLE NAME; HEIGHT AND UNIT
echo ''
log_debug 'Checking variable names, heights and units...'
for VAR in ${VARIABLES[@]}; do

  # skipping if pvap data is missing  
  if [ ${VAR//\"} == "pvap" ] && [ $(jq ".files.${VAR} | length" $1) == 0 ]; then
      log_debug "Skipping pvap"
      continue
  fi

  VAR_FILE_NAME=${TMP_DIR}'/'${VAR//\"}'_'${START_YEAR}'-'${END_YEAR}'.nc'

  X_SIZE=$(get_x_size ${VAR_FILE_NAME})
  Y_SIZE=$(get_y_size ${VAR_FILE_NAME})
  log_debug "Grid info: ${X_SIZE}x${Y_SIZE}"

  # grepping unit from attribute table of data file, check and convert if needed
  VAR_UNIT=$(cdo partab ${VAR_FILE_NAME} | grep "units" | grep -o -P '(?<=").*(?=")')
  VAR_NAME=$(cdo codetab ${VAR_FILE_NAME} | awk '{print $2;}')  ### Error if there are multiple variables -> needs to be tested
  VAR_HEIGHT=$(get_height ${VAR_FILE_NAME})

  log_debug "${VAR} [${VAR_UNIT}], HEIGHT: ${VAR_HEIGHT}M"

  if [ ! ${VAR_NAME} == ${VAR//\"} ]; then
    log_error "ERROR: Variable name in data (${VAR_NAME}) doesn't match variable name in data file name (${VAR})"
  fi

  # converting units if neccessary
  convert_unit ${VAR_FILE_NAME} ${VAR//\"}
done

### CALCULATION OF PVAP ###
if [ $(jq ".files.pvap | length" $1) == 0 ]; then
  log_debug "Calculating pvap"
  TASMAX_FILE=${TMP_DIR}'/tasmax_'${START_YEAR}'-'${END_YEAR}'.nc'
  TASMIN_FILE=${TMP_DIR}'/tasmin_'${START_YEAR}'-'${END_YEAR}'.nc'
  HURS_FILE=${TMP_DIR}'/hurs_'${START_YEAR}'-'${END_YEAR}'.nc'
  CALC_TMP_FILE=${TMP_DIR}'/tmp_calc_data.nc'

  ncks -A ${TASMAX_FILE} ${CALC_TMP_FILE}
  ncks -A ${TASMIN_FILE} ${CALC_TMP_FILE}
  ncks -A ${HURS_FILE} ${CALC_TMP_FILE}

  PVAP_MYEXP_FILE=$(dirname "$0")/helper/calc_pvap
  cdo -s -exprf,${PVAP_MYEXP_FILE} ${CALC_TMP_FILE} ${TMP_DIR}'/pvap_'${START_YEAR}'-'${END_YEAR}'.nc'
fi

### ALTITUDE ###
ALTITUDE_NAME=$(cdo codetab ${ALTI_FILE} | awk '{print $2;}')
ALTITUDE_UNIT=$(cdo partab ${ALTI_FILE} | grep "units" | grep -o -P '(?<=").*(?=")')
TMP_ALT_FILE=${TMP_DIR}'/'${ALTITUDE_NAME}'_'${START_YEAR}'-'${END_YEAR}'.nc'

log_debug "${ALTITUDE_NAME} [${ALTITUDE_UNIT}]"

case "${ALTITUDE_UNIT}" in
      "m")
        cdo -s -setmissval,nan -select,name=${ALTITUDE_NAME} ${ALTI_FILE} ${TMP_ALT_FILE}
        ;;
      *)
        log_error "ERROR: Altitude unit must be 'm'"
        exit 1
        ;;
esac

if [ ! ${REGION} == null ]; then
	cdo -s sellonlatbox,${REGION} ${TMP_ALT_FILE} 'regional_alt_data.nc'
	mv 'regional_alt_data.nc' ${TMP_ALT_FILE}
fi

##################################################################################################

echo ''
echo '### Calculating PET'
### LOOP OVER YEARS (COPY TO 1 INPUT-FILE, CALCULATION) ###
COUNTER=0
for ((YEAR=$START_YEAR;YEAR<END_YEAR+1;YEAR+=1)); do
{
  (( COUNTER++ ))
  log_debug ${YEAR}'/'${COUNTER}
  (
    # create temp. directory
    TMP_DIR_YEAR=${TMP_DIR}'/'${YEAR}
    [[ -d ${TMP_DIR_YEAR} ]] || mkdir  ${TMP_DIR_YEAR}

    # copy data in one file
    INPUT_FILE=${TMP_DIR_YEAR}'/data_'${YEAR}'.nc'
    TMP_FILE=${TMP_DIR_YEAR}'/tmp.nc'

    # TMAX
    cdo -s -selyear,${YEAR} ${TMP_DIR}'/tasmax_'${START_YEAR}'-'${END_YEAR}'.nc' ${INPUT_FILE}
    # TMIN
    cdo -s -selyear,${YEAR} ${TMP_DIR}'/tasmin_'${START_YEAR}'-'${END_YEAR}'.nc' ${TMP_FILE}
    ncks -A ${TMP_FILE} ${INPUT_FILE}
    # WIND
    cdo -s -selyear,${YEAR} ${TMP_DIR}'/sfcWind_'${START_YEAR}'-'${END_YEAR}'.nc' ${TMP_FILE}
    ncks -A ${TMP_FILE} ${INPUT_FILE}

    # add Wind_height as new variable (required for log. windprofile ...m to 2 m)
    TMP_FILE2=${TMP_DIR_YEAR}'/tmp2.nc'
    WIND_HEIGHT=$(get_height ${TMP_FILE})
    X_SIZE=$(get_x_size ${TMP_FILE})
    Y_SIZE=$(get_y_size ${VAR_FILE_NAME})
    cdo -s -setmissval,nan -setattribute,${WIND_H_NAME}@units="m" -chname,sfcWind,${WIND_H_NAME} -setcindexbox,${WIND_HEIGHT},1,${X_SIZE},1,${Y_SIZE} ${TMP_FILE} ${TMP_FILE2}  # windspeed -> wind_height=10m

    #cdo(2) setcindexbox (Warning): First latitude index out of range, set to 1!
    #cdo(2) setcindexbox (Warning): Last longitude index out of range, set to 1!
    ncatted -O -a long_name,${WIND_H_NAME},o,c,${WIND_H_NAME} ${TMP_FILE2}      # set standart/long name
    ncatted -O -a standard_name,${WIND_H_NAME},o,c,${WIND_H_NAME} ${TMP_FILE2}
    ncks -A ${TMP_FILE2} ${INPUT_FILE}

    # VAP
    cdo -s -selyear,${YEAR} ${TMP_DIR}'/pvap_'${START_YEAR}'-'${END_YEAR}'.nc' ${TMP_FILE}
    ncks -A ${TMP_FILE} ${INPUT_FILE}

     # SOL
     cdo -s -selyear,${YEAR} ${TMP_DIR}'/rsds_'${START_YEAR}'-'${END_YEAR}'.nc' ${TMP_FILE}
     ncks -A ${TMP_FILE} ${INPUT_FILE}

     # ALTITUDE
     ncks -A ${TMP_DIR}'/orog_'${START_YEAR}'-'${END_YEAR}'.nc' ${INPUT_FILE}
      # ncks: WARNING nco_cpy_var_dfn_trv() reports variable "lat" output type = NC_DOUBLE does not equal input type = NC_FLOAT. This is legal yet usually ill-advised when appending variables (i.e., with -A). Writing values into slots created for a different type is begging for trouble (e.g., data corruption, truncation, gingivitis).
      # lat, lon, rlat, rlon HSURF (float),  tasmax... (double)
    rm ${TMP_FILE}
    rm ${TMP_FILE2}

    MYEXP_FILE=$(dirname "$0")/helper/myexpr
    cdo -s exprf,${MYEXP_FILE} ${INPUT_FILE} ${TMP_DIR_YEAR}'/res_'${YEAR}'.nc'
    if [ ! -f ${TMP_DIR_YEAR}'/res_'${YEAR}'.nc' ]; then
        log_error 'Error create: '${TMP_DIR_YEAR}'/res_'${YEAR}'.nc'
        exit 1
    fi

    cdo -s -setmissval,-1.0E20 -chname,ET0,${PET_NAME} -selname,ET0 ${TMP_DIR_YEAR}'/res_'${YEAR}'.nc' ${TMP_DIR}'/'${PET_NAME}'_'${YEAR}'_tmp.nc'
    ncatted -O -a units,${PET_NAME},c,c,"""${PET_UNIT}""" ${TMP_DIR}'/'${PET_NAME}'_'${YEAR}'_tmp.nc'
    ncatted -O -a long_name,${PET_NAME},c,c,"""${PET_LNAME}""" ${TMP_DIR}'/'${PET_NAME}'_'${YEAR}'_tmp.nc'
    ncatted -O -a standard_name,${PET_NAME},c,c,"""${PET_SNAME}""" ${TMP_DIR}'/'${PET_NAME}'_'${YEAR}'_tmp.nc'
    ncatted -O -a coordinates,${PET_NAME},m,c,"lat lon" ${TMP_DIR}'/'${PET_NAME}'_'${YEAR}'_tmp.nc'       #coordinates = "height lat lon"; -> :coordinates = "lat lon";
    ncatted -O -a bounds,lat,d,, ${TMP_DIR}'/'${PET_NAME}'_'${YEAR}'_tmp.nc' # remove comment "lat_bnds"
    ncatted -O -a bounds,lon,d,, ${TMP_DIR}'/'${PET_NAME}'_'${YEAR}'_tmp.nc' # remove comment "lon_bnds"
    ncks -C -O -x -v lat_bnds,lon_bnds,height_2m ${TMP_DIR}'/'${PET_NAME}'_'${YEAR}'_tmp.nc' ${TMP_DIR}'/'${PET_NAME}'_'${YEAR}'.nc'
    # :cell_methods = "time: sum";
    rm ${TMP_DIR}'/'${PET_NAME}'_'${YEAR}'_tmp.nc'

    # copy to outour dir. ###
    nccopy -d 1 -s ${TMP_DIR}'/'${PET_NAME}'_'${YEAR}'.nc' ${OUTPUT_DIR_VAR}'/'${PET_NAME}'_'${YEAR}'010100-'$((${YEAR}+1))'010100.ncz'
    chmod 664 ${OUTPUT_DIR_VAR}'/'${PET_NAME}'_'${YEAR}'010100-'$((${YEAR}+1))'010100.ncz'
  ) &
  if [[ ${COUNTER} -ge ${MAXPP} ]]
  then
    COUNTER=0
    wait
  fi
}
done
wait
cd ${OUTPUT_DIR_VAR} && zip -r ${OUTPUT_DIR_VAR}/01_produced_data.zip . && cd -
echo '*** Calculation done ***'
